using System;
using System.IO;
using System.Threading;

using KnownObjects;

namespace Server
{
	/// <summary>
	/// Downloads a content.
	/// </summary>
	public class Downloader
	{
		/// <summary>
		/// Constructs Downloader instance and start a thread.
		/// </summary>
		/// <param name="iDownloadableContent"></param>
		/// <param name="totalSize"></param>
		/// <param name="iDownloadConfirmation"></param>
		public Downloader(IDownloadableContent iDownloadableContent, int totalSize, IDownloadConfirmation iDownloadConfirmation)
		{
			this.IDownloadableContent = iDownloadableContent;
			this.TotalSize = totalSize;
			this.IDownloadConfirmation = iDownloadConfirmation;

			Thread thread = new Thread(new ThreadStart(this.DownloadContent));
			thread.IsBackground = true;
			thread.Start();
		}

		public IDownloadableContent IDownloadableContent;
		public int TotalSize;
		public IDownloadConfirmation IDownloadConfirmation;

		public const int PORTION_SIZE = 65500;
		public const string DestinationDirectory = @"C:\TMP\";

		private delegate void SendPercentAsyncDelegate(int percent);

		/// <summary>
		/// Downloads a content.
		/// </summary>
		public void DownloadContent()
		{
			try
			{
				string fileName = DestinationDirectory + Guid.NewGuid().ToString("N");
				Stream outputStream = File.Create(fileName);

				for ( int bytesDownloaded = 0; bytesDownloaded < this.TotalSize; )
				{
					// request the next portion
					int sizeToRead = Math.Min(PORTION_SIZE, this.TotalSize - bytesDownloaded);

					// this is synchronous call that will be failed if content provider
					// will not reply for the specified time-out 
					// (Belikov.GenuineChannels.Messaging.MessageManager.TimeSpanWaitReplyTimeOut value)
					// 120 secs by default
					byte[] content = this.IDownloadableContent.GetNextPortion(sizeToRead);

					// save it into the file
					outputStream.Write(content, 0, content.Length);

					// update counters
					bytesDownloaded += content.Length;

					int percent = bytesDownloaded * 100 / this.TotalSize;

//					// calculate downloaded percent and send it in async call
//					// actually one-way call is good for this, but such features as
//					// ClientSession, remote peer info will not work
//					AsyncCallback responseCallback = new AsyncCallback(this.AsyncCallback);
//					SendPercentAsyncDelegate theCall = new SendPercentAsyncDelegate(this.IDownloadConfirmation.Progress);
//					theCall.BeginInvoke(percent, responseCallback, theCall);

					// I would recommend to use one-way method for notifying of progress
					// unless you need ClientSession or host info.
					this.IDownloadConfirmation.OneWayProgress(percent);
				}

				this.IDownloadConfirmation.Downloaded();
			}
			catch(Exception ex)
			{
				Console.WriteLine("An exception occuried while uploading a content: {0}.", ex.Message);
			}
		}

		/// <summary>
		/// Catches async replies from the remote host.
		/// Ignores results and exceptions.
		/// </summary>
		/// <param name="ar"></param>
		public void AsyncCallback(IAsyncResult ar)
		{
			try
			{
				SendPercentAsyncDelegate theCall = (SendPercentAsyncDelegate) ar.AsyncState;
				theCall.EndInvoke(ar);
			}
			catch(Exception)
			{
			}
		}
	}
}
